@@ -552,7 +552,7 @@ class KeyedList (list):
     def append(self, e):
         """Append element to list and record its key."""
         if e.name in self.elements:
-            raise ValueError('Element name %s already exists.' % e.name)
+            raise ValueError('Element name %s already exists.' % (e.name,))
         list.append(self, e)
         self.elements[e.name] = e
 
@@ -665,7 +665,7 @@ class CatalogTable (NodeConfigAclBinding):
             raise ValueError('column %s object is not from this table object' % (column,))
         elif column in self.column_definitions.elements:
             return self.column_definitions[column]
-        raise ValueError('value %s does not name a defined column in this table' % (c,))
+        raise ValueError('value %s does not name a defined column in this table' % (column,))
 
     def key_by_columns(self, unique_columns, raise_nomatch=True):
         """Return key from self.keys with matching unique columns.
@@ -912,7 +912,7 @@ class CatalogKey (NodeConfig):
         try:
             self.constraint_schema, self.constraint_name = _constraint_name_parts(self, key_doc)
         except ValueError:
-            self.constraint_schema, self.constraint_name = None, hash(self)
+            self.constraint_schema, self.constraint_name = None, str(hash(self))
         self.unique_columns = KeyedList([
             table.column_definitions[cname]
             for cname in key_doc['unique_columns']
@@ -990,7 +990,7 @@ class CatalogForeignKey (NodeConfigAclBinding):
         try:
             self.constraint_schema, self.constraint_name = _constraint_name_parts(self, fkey_doc)
         except ValueError:
-            self.constraint_schema, self.constraint_name = None, hash(self)
+            self.constraint_schema, self.constraint_name = None, str(hash(self))
         if self.constraint_schema:
             self.constraint_schema._fkeys[self.constraint_name] = self
         else:
